/************************************************************************
*									*
* The XOLDT System							*
* Author: David S. Warren						*
* V1.0 Copyright SUNY Stony Brook, 1992					*
* This code is distributed WITHOUT ANY WARRANTY.			*
*									*
************************************************************************/

/* This does constructive negation using xoldt. It uses the
constraint handling within XOLDT, re-defining xand and xcomp. The
code was taken from the network as posted by Mark Johnson, and
modified. Many of the following comments and examples are taken
verbatim from that posting. */ 

%This file contains code modified from a metainterpreter for negation 
% that runs in Quintus Prolog. It was posted to the network by Mark 
% Johnson. 
%
%The treatment of negation is a generalization of the standard
%negation as failure.
%
%Consider a (positive) goal p(X). The standard resolution proof of
%p(X) from a set of program axioms A yields a sequence of answer 
%substitutions T1, .., Tn, such that A |= p(T) for any T that is
%subsumed by some Ti in T1, .., Tn.
%
%Assuming the completion, A |= ~p(T) iff for all answer substitutions
%Ti from a resolution proof of p(X), T is not subsumed by any such Ti.
%The metainterpreter below passes around a list of such anti-subsumption
%constraints, and continually checks them to see if they are violated.
%
%
%Here are some samples.
%
%%% X is not a. Cs is a list of antisubsumption constraints.
%| ?- (~X=a) <- Cs.
%
%X = _3119,
%Cs = [[X]/>[a]] ;	% X must not be subsumed by a.
%
%no
%%% This is diff(X,Y)
%| ?- (~X=Y) <- Cs.
%
%X = _3119,
%Y = _3136,
%Cs = [[X,Y]/>[_3711,_3711]] ; % The pair X,Y must not be subsumed by U,U
%
%no
%%% X is not a member of [a,b,c]
%| ?- (~member(X,[a,b,c])) <- Cs.
%
%X = _3129,
%Cs = [[X]/>[a],[X]/>[b],[X]/>[c]] ; % X must not be subsumed by a, b, or c.
%
%no
%%% X is not a member of [a,Y]
%| ?- (~member(X,[a,Y])) <- Cs.
%
%X = _3129,
%Y = _3164,
%Cs = [[X,Y]/>[a,_3815],[X,Y]/>[_3822,_3822]] ; % X must not be subsumed by a,
%	 				% and X,Y must not be subsumed by U,U
%no
%
% This next goal uses double negation and existential quantification.
% Literally it reads: L is a list, and it is not the case that there
% exists an X that is a member of L and not equal to a, i.e. every
% member of L is a.
%
%| ?- list(L),(~X^(member(X,L),~X=a)) <- Cs.
%
%L = Cs = [],
%X = _3160 ;
%
%L = [a],
%X = _3160,
%Cs = [] ;
%
%L = [a,a],
%X = _3160,
%Cs = [] ;
%
%L = [a,a,a],
%X = _3160,
%Cs = [] ;
%
%L = [a,a,a,a],
%X = _3160,
%Cs = [] 
%
:- op(950, fx, ~). % Negation operator
:- op(700, xfx, />). % Anti-subsumption constraint
:- op(800, xfx, &).

:- ensure_loaded(xoldt).
:- ensure_loaded(library(lists)).
:- ensure_loaded(freevars).
:- ensure_loaded(subsumes).

% Define xand: Ci is a list of anti-subsumption constraints.
xand(C1,C2,C) :- 
	append(C1,C2,C3), check_negs(C3,C).

% Define ~
:- notable (~)/1.
~G :- free_variables(G,[],[],FVs),
	findall(FVs&Cs,G<-Cs,NewCs),
	negate(NewCs,FVs,NegCs),
	xconstraint(NegCs).


% define duplicate constraints
%	(simple variant might suffice.)
xcomp(le,[H1|B1],[H2|B2]) :- \+ \+ (
	numbervars(H1,0,N),numbervars(H2,0,N),H1==H2,
	is_subset(B1,B2)
	).

is_subset([],_).
is_subset([X|L1],L2) :- member(X1,L2),variant(X,X1),!,is_subset(L1,L2).


% check_negs(CIns, COuts)
% Checks anti-subsumption constraints.
check_negs([], []).
check_negs([T0/>T1|C0s], C1s) :-
	\+ T0=T1, % T1 can never subsume T0
	!, % so delete this constraint
	check_negs(C0s, C1s).
check_negs([T0/>T1|_], _) :-
	subsumes(T1, T0), % T1 subsumes T0, so fail
	!,
	fail.
check_negs([C|C0s], [C|C1s]) :-
	check_negs(C0s, C1s).

% negate(ToNegate, FVs, Cs)
% Converts the output of the findall of a negated goal
% 	into anti-subsumption constraints.
negate([], _, []).
negate([Vals&_|NewCs], FVs, [FVs/>Vals|Cs]) :-
	negate(NewCs, FVs, Cs).
negate([FVs&Negs|NewCs], FVs, Cs) :-
	member(FV1s/>FV1s, Negs),
	negate(NewCs, FVs, Cs).


/*********************************************************************
 * t.pl Example programs for negation metainterpreter m1.
 */

:- op(700, xfx, /=).


:- notable (/=)/2.
% Inequality, i.e. the diff of Prolog II and Sicstus
X/=Y :- ~X=Y.


% list(L) iff L is a proper list
list([]).
list([_|Xs]) :- list(Xs).

% list_of(E, Es) iff Es is a proper list, every member of which is E.
%| ?- list(L),list_of(a, L) <- Cs.
%
%L = Cs = [] ;
%
%L = [a],
%Cs = [] ;
%
%L = [a,a],
%Cs = [] ;
%
%L = [a,a,a],
%Cs = [] 
% ...
:- notable list_of/2.
list_of(E, Es) :- 
	list(Es),
	~ X^(member(X,Es),E/=X).

:- notable every/3.
% every(X,P,Q) iff forall X, P(X) implies Q(X) which is equivalent
%	to ~ exists X, P(X) and ~ Q(X)
every(X,P,Q) :- ~ X^(P,~Q).

:- notable duplicated/2.
% duplicated(E,L) iff E appears at least twice in L.
duplicated(E, L) :- 
	append(_, [E|Es], L),
	member(E, Es).

% unique_list(L) iff no element appears twice in L,
% i.e. there does not exist an E in L duplicated.
%| ?- list(L),unique_list(L)<-Cs.
%
%L = Cs = [] ;
%
%L = [_3335],
%Cs = [] ;
%
%L = [_3335,_3337],
%Cs = [[_3335,_3337]/>[_4745,_4745]]
%	...
:- notable unique_list/1.
unique_list(L) :- 
	list(L),
	~ E^duplicated(E, L).

% all_duplicates(L) iff every element in L appears at
% least once.
%| ?- all_duplicates([a,b,U,V])<-Cs.
%
%U = a,
%V = b,
%Cs = [] ;
%
%U = b,
%V = a,
%Cs = [] ;
%
%no

:- table all_duplicates/1.
all_duplicates(L) :- 
	list(L),
	every(E, member(E,L), duplicated(E,L)).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% To show that loops are handled.

:- table( path/2).
:- table( edge/2).
:- table( red/1).
path(X,Y) :- edge(X,Y).
path(X,Y) :- path(X,Z), edge(Z,Y).

edge(a,b).
edge(b,a).
edge(a,c).
red(c).

% There is no (nonempty) path to a red node.
:- table(inaccred/1).
inaccred(X) :- ~path(X,Y),red(Y).

% which is true for any node other than a or b.
%| ?- inaccred(X)<-DEL.
%
%X = _3119,
%DEL = [[X,c]/>[a,c],[X,c]/>[b,c]] ;
%
%no

