/************************************************************************
*									*
* The XOLDT System							*
* Author: David S. Warren						*
* V1.0 Copyright SUNY Stony Brook, 1992					*
* This code is distributed WITHOUT ANY WARRANTY.			*
*									*
************************************************************************/

/* This module (xoldt_load) supports consulting and compiling of
files for oldt evaluation. This module defines several predicates.
The main one is do_term_expansion/2 which should be bound to
user:term_expansion/2 so it will be called during compilation.

Once this file is loaded and term_expansion changed, clauses will be
intercepted and modified for oldt evaluation. 

For each tabled predicate e.g. p/3, the rule:
	Mod:'xoldt$$rule'(p(X,Y,Z),B) :- 'xoldt$p'(X,Y,Z,B). 
is asserted into the module in which it is to be defined. Also a rule: 
	p(X,Y,Z) :- xoldt:xoldt(Mod:p(X,Y,Z)). 
is asserted. and each rule such as
	p(a,X,Y) :- q(X,Z),!,r(Z,Y),s(Y). 
for q prolog and r,s tabled, is transformed into
	'xoldt$p'(a,X,Y,(r(Z,Y),s(Y)) :- q(X,Z),!. 
and returned, by term-expansion.

The idea here is to do a little optimization so that Prolog's
underlying indexing can be used. The other thing is to evaluate
``guards'' in Prolog. So for a clause whose first ``chunk'' contains
only builtins or prolog-evaluated subgoals, these subgoals are
evaluated in Prolog directly, and not tabled. It allows simple
testing to be done in Prolog. For example, ET will not accept cuts
in general, but if one follows a sequence of builtins, it can be done
in Prolog and not mess up the tabling. E.g. a clause such as: 
	p(X,Y) :- X<7, !, q(X,Y). 
is accepted for a tabled predicate, and the test and cut is done
during clause selection, so never enters the tabling. */

/* This code doesn NOT support modules fully. At this time, there is
bug in prolog_load_context so the embedded use_module's don't work
right. Even if that did work, tabling across module boundaries
doesn't work. The only way to call a tabled predicate from another
module within the meta-interpreter is to explicitly attach the
module name to the call: e.g. ...,mod2:path(X,Y),...
*/

:- module(xoldt_load, []).


:- dynamic xoldt_expanding/0.
:- dynamic (default)/2.
% o_preds(Pred,Mod,File)
:- dynamic o_preds/3.
% intercepting(Mod,File)
:- dynamic intercepting/2.	
% o_type(Pred,Type,File)
:- dynamic o_type/3.

/*** BUG, this doesn't handle embedded compiles quite right ***/

%configuration junk to support (partially) earlier versions of qp 
:- use_module(library(lists)).

% CHANGED GvN!
prolog_load_context(module,Module) :-
	prolog_flag(typein_module,Module).
prolog_load_context(file,File) :-
	findall(File,current_stream(File,read,_),Files),
	last(Files,File).

% imported from library(lists) in Sicstus
%last([F],F) :- !.
%last([_|L],F) :- last(L,F).


do_term_expansion(Clause,New_clause) :-
	prolog_load_context(file,File), !,
	%write(intercepted(File,Clause)),nl,	%
	(intercepting(Mod,File)
	 ->	%write(intercepting(Mod,Clause)),nl,	%
		expand(Clause,Mod,File,New_clause)
		 %,write(returning(New_clause)),nl
	 ; table_command(Clause)
	 ->	prolog_load_context(module,Mod),
		%write(init_and_exp(Mod,Clause)),nl,	%
		init(Mod,File),
		expand(Clause,Mod,File,New_clause)
		 %,write(returning(New_clause)),nl
	).

table_command(:-(table(_))).
table_command(:-(notable(_))).
table_command(:-(prolog(_))).
table_command(:-(default(_))).

init(Mod,File) :- 
	asserta(intercepting(Mod,File)),
	retractall(Mod:'xoldt$$rule'('xoldt$$table'(_),_)),
	assertz(Mod:'xoldt$$rule'('xoldt$table'(X),X)),
	retractall(Mod:'xoldt$$type'('xoldt$table'(_),_)),
	assertz(Mod:'xoldt$$type'('xoldt$table'(_),(table))),
	clean_old(Mod,File),
	add_xmeta(Mod).

add_xmeta(Mod) :-
	xoldt_builtin(P),
	retractall(Mod:'xoldt$$type'(P,_)),
	assertz(Mod:'xoldt$$type'(P,xmeta)),
	fail.
add_xmeta(_Mod).

clean_old(Mod,File) :-
	o_preds(Spec,Mod,File),
	retractall(Mod:'xoldt$$rule'(Spec,_)),
	retractall(Mod:'xoldt$$type'(Spec,_)),
	fail.
clean_old(Mod,File) :- 
	retractall(o_preds(_,Mod,File)).


expand(end_of_file,_Mod,File,_) :- !,
	retractall(default(_,File)),
	retract(intercepting(_,File)),
	retractall(o_type(_,_,File)),
	fail.
expand(:-(Com),Mod,File,:-(New_com)) :- !, 
	expand_command(Com,Mod,File,New_com).
expand((H-->B),Mod,File,New_clause) :- !,
	\+ xoldt_expanding,
	assert(xoldt_expanding),
	expand_term((H-->B),Clause1),
	retractall(xoldt_expanding),
	expand(Clause1,Mod,File,New_clause).
expand(Clause,Mod,File,New_clause) :-
	(Clause = (Head:-Body) -> true ; Head = Clause, Body=true),
	\+ prolog_pred(Head,Mod,File),
	process_term(Head,Body,Mod,File,New_clause)
	.	%,write(compiled(New_clause)),nl.

expand_command(prolog((A,B)),Mod,File,true) :- !,
	expand_command_prolog(A,Mod,File),
	expand_command(prolog(B),Mod,File,true).
expand_command(prolog(A),Mod,File,true) :- 
	expand_command_prolog(A,Mod,File).
expand_command(table((A,B)),Mod,File,true) :- !,
	expand_command_table(A,Mod,File),
	expand_command(table(B),Mod,File,true).
expand_command(table(A),Mod,File,true) :- 
	expand_command_table(A,Mod,File).
expand_command(notable((A,B)),Mod,File,true) :- !,
	expand_command_notable(A,Mod,File),
	expand_command(notable(B),Mod,File,true).
expand_command(notable(A),Mod,File,true) :- 
	expand_command_notable(A,Mod,File).
expand_command(default(Type),_Mod,File,true) :-
	retractall(default(_,File)),
	assert(default(Type,File)).

expand_command_prolog(A,Mod,File) :-
	(A=P/N -> functor(T,P,N) ; T = A),
	((o_type(T,Ot,_),Ot \== (prolog))
	 ->	write('ERROR: '),write(P/N),write(' assumed to be of type '),
		write(Ot),nl,tab(7),write('But later declared prolog.'),nl
	 ;	assertz(o_type(T,(prolog),File)),
		retractall(Mod:'xoldt$$type'(T,_))
	).

expand_command_table(A,Mod,File) :-
	(A=P/N -> functor(T,P,N) ; T = A),
	(predicate_property(T,built_in)
	 ->	write('ERROR: Cannot table a built_in.')
	 ; (o_type(T,Ot,_),Ot \== (table))
	 ->	write('ERROR: '),write(P/N),write(' assumed to be of type '),
		write(Ot),nl,tab(7),write('But later declared table.'),nl
	 ;	assertz(o_type(T,(table),File)),
		retractall(Mod:'xoldt$$type'(T,_)),
		assertz(Mod:'xoldt$$type'(T,(table)))
	).

expand_command_notable(A,Mod,File) :-
	(A=P/N -> functor(T,P,N) ; T = A),
	(predicate_property(T,built_in)
	 ->	write('ERROR: Cannot table a built_in.')
	 ; (o_type(T,Ot,_),Ot \== (notable))
	 ->	write('ERROR: '),write(P/N),write(' assumed to be of type '),
		write(Ot),nl,tab(7),write('But later declared notable.'),nl
	 ;	assertz(o_type(T,(notable),File)),
		retractall(Mod:'xoldt$$type'(T,_)),
		assertz(Mod:'xoldt$$type'(T,(notable)))
	).


process_term(H,B,Mod,File,New_clause) :-
	functor(H,P,A),
	name(P,Pl),
	name(Npred,[120,111,108,100,116,36|Pl]),	% 'xoldt$'
	functor(Gen,P,A),
	flatten(Gen,Gbod,Npred,Bcall),
	(o_preds(Gen,Mod,File)
	 ->	true
	 ;	((o_preds(Gen,Mod,Ofile),File \== Ofile)
		 ->	write('Procedure '),write(P/A),
			write(' is being redefined in a different file - '),
			nl,
			tab(4),write('Previous file: '),write(Ofile),nl,
			tab(4),write('New file: '),write(File),nl,
			retractall(o_preds(Gen,Mod,Ofile))
		 ;	true
		),
		abolish(Mod:P/A),
		Mod:assertz((Gen :- xoldt:xoldt(Mod:Gen))),
		Mod:assertz(('xoldt$$rule'(Gen,Gbod) :- Bcall)),
		assertz(o_preds(Gen,Mod,File))
	),
	process_rule(H,B,Mod,File,Npred,New_clause).
	

prolog_pred(T,_Mod,_File) :- xoldt_builtin(T),!,fail.
prolog_pred(T,_Mod,_File) :- predicate_property(T,built_in),!.
prolog_pred(T,_Mod,_File) :- o_type(T,(prolog),_),!.
prolog_pred(T,_Mod,_File) :- o_type(T,(table),_),!,fail.
prolog_pred(T,_Mod,_File) :- o_type(T,(notable),_),!,fail.
prolog_pred(T,Mod,File) :- default((table),File),!,
	functor(T,P,N),functor(T1,P,N),
	assertz(o_type(T1,(table),File)),
	assertz(Mod:'xoldt$$type'(T1,(table))),fail.
prolog_pred(T,Mod,File) :- default((notable),File),!,
	functor(T,P,N),functor(T1,P,N),
	assertz(o_type(T1,(notable),File)),
	assertz(Mod:'xoldt$$type'(T1,(notable))),fail.
prolog_pred(T,_Mod,File) :- 
	functor(T,P,N),functor(T1,P,N),
	assertz(o_type(T1,(prolog),File)).

% e.g. flatten(p(A,B),C,p2,p2(A,B,C)).
flatten(Term1,Term2,Npred,Flat) :-
	Term1 =.. [_|Term1args],
	append(Term1args,[Term2],Flatargs),
	Flat =.. [Npred|Flatargs].


process_rule(Head,Body,Mod,File,Npred,(Clausehd:-Clausebd)) :-
	split_init_prolog(Body,Mod,File,Prolog,Et),
	Head =.. [_|Form_pars],
	append(Form_pars,[Et],Clauseargs),
	Clausehd =.. [Npred|Clauseargs],
	Clausebd = Prolog.
	

split_init_prolog((A,B),Mod,File,Pro,Et) :- !,
	(prolog_pred(A,Mod,File)
	 ->	split_init_prolog1(A,B,Mod,File,Pro,Et)
	 ;	Pro = true, Et = (A,B), check_rest(Et,Mod,File)
	).
split_init_prolog(A,Mod,File,Pro,Et) :-
	(prolog_pred(A,Mod,File)
	 ->	Pro = A, Et = true
	 ;	Pro = true, Et = A, check_rest(Et,Mod,File)
	).

split_init_prolog1(A,(B,C),Mod,File,Pro,Et) :- !,
	(prolog_pred(B,Mod,File)
	 ->	split_init_prolog1(B,C,Mod,File,Pror,Etr),
		Pro = (A,Pror), Et = Etr
	 ;	Pro = A, Et = (B,C), check_rest(Et,Mod,File)
	).
split_init_prolog1(A,B,Mod,File,Pro,Et) :-
	(prolog_pred(B,Mod,File)
	 ->	Pro = (A,B), Et = true
	 ;	Pro = A, Et = B, check_rest(Et,Mod,File)
	).

check_rest((A,B),Mod,File) :- !,
	check_l(A,Mod,File),
	check_rest(B,Mod,File).
check_rest(A,Mod,File) :- !,
	check_l(A,Mod,File).


check_l((If->A),Mod,File) :- 
	(prolog_pred(If,Mod,File)
	 ->	true
	 ;	write('ERROR: Test in -> must be prolog: '),
		write(If),write(' is not!'),nl
	),
	check_l(A,Mod,File).
check_l((If->A;B),Mod,File) :- !,
	(prolog_pred(If,Mod,File)
	 ->	true
	 ;	write('ERROR: Test in -> must be prolog: '),
		write(If),write(' is not!'),nl
	),
	check_l(A,Mod,File),
	check_l(B,Mod,File).
check_l((A;B),Mod,File) :- 
	check_l(A,Mod,File),
	check_l(B,Mod,File).
check_l((A,B),Mod,File) :-
	check_l(A,Mod,File),
	check_l(B,Mod,File).
check_l(A,Mod,File) :-
	(A == true -> true
	 ; A == ! -> write('Illegal use of cut!'),nl
	 ; prolog_pred(A,Mod,File)
	 ->	true
	 ;	true
	).

%xoldt_builtin((_->_)).
%xoldt_builtin((_->_;_)) :- !.
xoldt_builtin((_;_)).
xoldt_builtin((_,_)).
xoldt_builtin(findall(_,_,_)).
xoldt_builtin(clause(_,_)).
xoldt_builtin(call(_)).
xoldt_builtin(^(_,_)).
xoldt_builtin(<-(_,_)).
xoldt_builtin(xoldt(_)).
xoldt_builtin(xconstraint(_)).
xoldt_builtin(xgetconstraint(_)).
